import itertools
import logging
from typing import List, Optional

from a_sync import a_sync, cgather, igather
from a_sync.executor import AsyncExecutor
from async_lru import alru_cache
from brownie.network.event import _EventItem
from eth_typing import HexStr
from eth_utils.toolz import concat
from evmspec.data import Address, HexBytes32, uint
from evmspec.structs.log import Topic
from hexbytes import HexBytes
from msgspec import DecodeError, ValidationError, json, msgpack
from pony.orm import commit, db_session, select
from pony.orm.core import Query

from y import convert
from y._db.common import DiskCache, default_filter_threads, enc_hook, make_executor
from y._db.decorators import db_session_cached, db_session_retry_locked, retry_locked
from y._db.entities import Block, Hashes, LogCacheInfo, LogTopic
from y._db.entities import Log as DbLog
from y._db.log import Log
from y._db.utils._ep import _get_get_block
from y._db.utils.bulk import insert as _bulk_insert
from y.constants import CHAINID

logger = logging.getLogger(__name__)

LOG_COLS = (
    "block_chain",
    "block_number",
    "tx",
    "log_index",
    "address",
    "topic0",
    "topic1",
    "topic2",
    "topic3",
    "raw",
)

_BLOCK_COLS = "chain", "number"
_BLOCK_COLS_EXTENDED = "chain", "number", "classtype"

_topic_executor = make_executor(4, 8, "ypricemagic db executor [topic]")
_hash_executor = make_executor(4, 8, "ypricemagic db executor [hash]")

_get_log_cache_info = LogCacheInfo.get
_get_log_topic = LogTopic.get
_get_hash = Hashes.get

_encode_generic = json.Encoder().encode
_encode_log = json.Encoder(enc_hook=enc_hook).encode


def _decode_hook_unsafe(typ, obj):
    """This decode hook does NOT ensure addresses are checksummed. They must be stored that way."""
    try:
        if issubclass(typ, uint):
            if isinstance(obj, int):
                return int.__new__(typ, obj)
            elif isinstance(obj, str):
                return typ(obj, 16)
        elif issubclass(typ, HexBytes):
            return typ(obj)
        elif typ is Address:
            # We have to put the 0x back on since we remove it for db storage.
            # The safe decode hook does that for us but the unsafe hook skips all validation.
            return str.__new__(Address, f"0x{obj}")
    except (TypeError, ValueError, ValidationError) as e:
        raise Exception(e, typ, obj) from e
    raise NotImplementedError(typ, obj)


_json_decode_log = json.Decoder(type=Log, dec_hook=_decode_hook_unsafe).decode
_msgpack_decode_log = msgpack.Decoder(type=Log, dec_hook=_decode_hook_unsafe).decode


def _decode_log(data: bytes) -> Log:
    try:
        # more recent versions of ypm store the logs in messagepack format
        return _msgpack_decode_log(data)
    except DecodeError:
        # but ypm can still work with logs stored in the legacy json format
        log = _json_decode_log(data)
        # we just update them to the new format silently
        tx_hash_dbid = _get_hash(hash=log.transactionHash.hex()[2:]).dbid
        DbLog[CHAINID, log.blockNumber, tx_hash_dbid, log.logIndex].raw = _encode_log(log)
        # and you're good to go
        return log


async def _prepare_log(log: Log) -> tuple:
    """
    Prepare a log for insertion into the database.

    This function gathers database IDs for the transaction hash, address, and topics
    of a given log. It then encodes the log as JSON using the `enc_hook` for special types.

    Args:
        log: The log entry to prepare.

    Returns:
        A tuple containing the prepared log parameters.

    Examples:
        >>> log = Log(transactionHash=HexBytes('0x1234'), address='0x...', topics=['0x...'], blockNumber=123, logIndex=0)
        >>> prepared_log = await _prepare_log(log)
        >>> print(prepared_log)

    See Also:
        - :func:`get_hash_dbid`
        - :func:`get_topic_dbid`
        - :func:`enc_hook`
    """
    transaction_dbid, address_dbid, topic_dbids = await cgather(
        get_hash_dbid(log.transactionHash.hex()),
        get_hash_dbid(log.address),
        igather(map(get_topic_dbid, log.topics)),
    )
    topics = {
        f"topic{i}": topic_dbid for i, topic_dbid in itertools.zip_longest(range(4), topic_dbids)
    }
    params = {
        "block_chain": CHAINID,
        "block_number": log.blockNumber,
        "transaction": transaction_dbid,
        "log_index": log.logIndex,
        "address": address_dbid,
        **topics,
        "raw": _encode_log(Log(**log)),
    }
    return tuple(params.values())


_check_using_extended_db = lambda: "eth_portfolio" in _get_get_block().__module__


async def bulk_insert(logs: List[Log], executor: AsyncExecutor = default_filter_threads) -> None:
    if not logs:
        return

    submit = executor.submit

    # handle a conflict with eth-portfolio's extended db
    if _check_using_extended_db():
        blocks = tuple(
            (CHAINID, block, "BlockExtended") for block in {log.blockNumber for log in logs}
        )
        blocks_fut = submit(_bulk_insert, Block, _BLOCK_COLS_EXTENDED, blocks, sync=True)
    else:
        blocks = tuple((CHAINID, block) for block in {log.blockNumber for log in logs})
        blocks_fut = submit(_bulk_insert, Block, _BLOCK_COLS, blocks, sync=True)
    del blocks

    txhashes = (txhash.hex() for txhash in {log.transactionHash for log in logs})
    addresses = {log.address for log in logs}
    hashes = tuple((_remove_0x_prefix(hash),) for hash in itertools.chain(txhashes, addresses))
    hashes_fut = submit(_bulk_insert, Hashes, ("hash",), hashes, sync=True)
    del txhashes, addresses, hashes

    topics = set(concat(log.topics for log in logs))
    topics_fut = submit(
        _bulk_insert,
        LogTopic,
        ("topic",),
        tuple((_remove_0x_prefix(topic.strip()),) for topic in topics),
        sync=True,
    )
    del topics

    await cgather(blocks_fut, hashes_fut, topics_fut)

    await executor.run(
        _bulk_insert,
        DbLog,
        LOG_COLS,
        await igather(map(_prepare_log, logs)),
        sync=True,
    )


@a_sync(default="async", executor=_topic_executor, ram_cache_maxsize=None)
@db_session_cached
def get_topic_dbid(topic: Topic) -> int:
    topic = _remove_0x_prefix(topic.strip())
    entity = _get_log_topic(topic=topic)
    if entity is None:
        entity = LogTopic(topic=topic)
    return entity.dbid


@alru_cache(maxsize=10000, ttl=600)
async def get_hash_dbid(txhash: HexStr) -> int:
    return await _get_hash_dbid(txhash)


@a_sync(default="async", executor=_hash_executor)
@db_session_retry_locked
def _get_hash_dbid(hexstr: HexStr) -> int:
    if len(hexstr) == 42:
        hexstr = convert.to_address(hexstr)
    string = _remove_0x_prefix(hexstr)
    entity = _get_hash(hash=string)
    if entity is None:
        entity = Hashes(hash=string)
    return entity.dbid


def get_decoded(log: Log) -> Optional[_EventItem]:
    # TODO: load these in bulk
    log = DbLog[CHAINID, log.block_number, log.transaction_hash, log.log_index]
    if decoded := log.decoded:
        return _EventItem(
            decoded["name"], decoded["address"], decoded["event_data"], decoded["pos"]
        )


@db_session
@retry_locked
def set_decoded(log: Log, decoded: _EventItem) -> None:
    log = DbLog[CHAINID, log.block_number, log.transaction_hash, log.log_index]
    log.decoded = decoded


page_size = 100


class LogCache(DiskCache[Log, LogCacheInfo]):
    __slots__ = "addresses", "topics"

    def __init__(self, addresses, topics):
        self.addresses = addresses
        self.topics = topics

    def __repr__(self) -> str:
        string = f"{type(self).__name__}(addresses={self.addresses}"
        for topic in ("topic0", "topic1", "topic2", "topic3"):
            if value := getattr(self, topic):
                string += f", {topic}={value}"
        return f"{string})"

    @property
    def topic0(self) -> str:
        return self.topics[0] if self.topics else None

    @property
    def topic1(self) -> str:
        return self.topics[1] if self.topics and len(self.topics) > 1 else None

    @property
    def topic2(self) -> str:
        return self.topics[2] if self.topics and len(self.topics) > 2 else None

    @property
    def topic3(self) -> str:
        return self.topics[3] if self.topics and len(self.topics) > 3 else None

    def load_metadata(self) -> Optional[LogCacheInfo]:
        """Loads the cache metadata from the db."""
        if self.addresses:
            raise NotImplementedError(self.addresses)

        from y._db.utils import utils as db

        chain = db.get_chain(sync=True)
        # If we cached all of this topic0 with no filtering for all addresses
        if self.topic0 and (
            info := _get_log_cache_info(
                chain=chain,
                address="None",
                topics=_encode_generic([self.topic0]),
            )
        ):
            return info
        # If we cached these specific topics for all addresses
        elif self.topics and (
            info := _get_log_cache_info(
                chain=chain,
                address="None",
                topics=_encode_generic(self.topics),
            )
        ):
            return info

    def _is_cached_thru(self, from_block: int) -> int:
        from y._db.utils import utils as db

        if self.addresses:
            chain = db.get_chain(sync=True)
            infos: List[LogCacheInfo]
            if isinstance(self.addresses, str):
                infos = [
                    # If we cached all logs for this address...
                    _get_log_cache_info(
                        chain=chain,
                        address=self.addresses,
                        topics=_encode_generic(None),
                    )
                    # ... or we cached all logs for these specific topics for this address
                    or _get_log_cache_info(
                        chain=chain,
                        address=self.addresses,
                        topics=_encode_generic(self.topics),
                    )
                ]
            else:
                infos = [
                    # If we cached all logs for this address...
                    _get_log_cache_info(chain=chain, address=addr, topics=_encode_generic(None))
                    # ... or we cached all logs for these specific topics for this address
                    or _get_log_cache_info(
                        chain=chain, address=addr, topics=_encode_generic(self.topics)
                    )
                    for addr in self.addresses
                ]
            if all(info and from_block >= info.cached_from for info in infos):
                return min(info.cached_thru for info in infos)

        elif (info := self.load_metadata()) and from_block >= info.cached_from:
            return info.cached_thru
        return 0

    def _select(self, from_block: int, to_block: int) -> List[Log]:
        logger.info("executing select query for %s", self)
        try:
            return [_decode_log(log.raw) for log in self._get_query(from_block, to_block)]
        except ValidationError:
            results = []
            for log in self._get_query(from_block, to_block):
                try:
                    results.append(_decode_log(log.raw))
                except ValidationError as e:
                    raise ValueError(e, json.decode(log.raw)) from e
            return results

    def _get_query(self, from_block: int, to_block: int) -> Query:
        from y._db.utils import utils as db

        generator = (
            log
            for log in DbLog
            if log.block.chain == db.get_chain(sync=True)
            and log.block.number >= from_block
            and log.block.number <= to_block
        )

        generator = self._wrap_query_with_addresses(generator)

        for topic in [f"topic{i}" for i in range(4)]:
            generator = self._wrap_query_with_topic(generator, topic)

        query = (
            select(generator)
            .without_distinct()
            .order_by(lambda l: (l.block.number, l.tx.hash, l.log_index))
        )
        logger.debug(query.get_sql())
        return query

    def _set_metadata(self, from_block: int, done_thru: int) -> None:
        from y._db.utils import utils as db

        chain = db.get_chain(sync=True)
        encoded_topics = _encode_generic(self.topics or None)
        should_commit = False
        if self.addresses:
            addresses = self.addresses
            if isinstance(addresses, str):
                addresses = [addresses]
            for address in addresses:
                if e := _get_log_cache_info(chain=chain, address=address, topics=encoded_topics):
                    if from_block < e.cached_from:
                        e.cached_from = from_block
                        should_commit = True
                    if done_thru > e.cached_thru:
                        e.cached_thru = done_thru
                        should_commit = True
                else:
                    LogCacheInfo(
                        chain=chain,
                        address=address,
                        topics=encoded_topics,
                        cached_from=from_block,
                        cached_thru=done_thru,
                    )
                    should_commit = True
        elif info := _get_log_cache_info(
            chain=chain,
            address="None",
            topics=encoded_topics,
        ):
            if from_block < info.cached_from:
                info.cached_from = from_block
                should_commit = True
            if done_thru > info.cached_thru:
                info.cached_thru = done_thru
                should_commit = True
        else:
            LogCacheInfo(
                chain=chain,
                address="None",
                topics=encoded_topics,
                cached_from=from_block,
                cached_thru=done_thru,
            )
            should_commit = True
        if should_commit:
            commit()
            logger.debug("cached %s %s thru %s", self.addresses, self.topics, done_thru)
        return

    def _wrap_query_with_addresses(self, generator) -> Query:
        if not (addresses := self.addresses):
            return generator
        elif isinstance(addresses, str):
            address = convert.to_address(addresses)[2:]
            return (log for log in generator if log.address.hash == address)
        addresses = tuple(convert.to_address(address)[2:] for address in addresses)
        return (log for log in generator if log.address.hash in addresses)

    def _wrap_query_with_topic(self, generator, topic_id: str) -> Query:
        if not (topic_or_topics := getattr(self, topic_id)):
            return generator

        if isinstance(topic_or_topics, (bytes, str)):
            topic = topic_or_topics
            topic = _remove_0x_prefix(HexBytes32(topic).strip())
            return (log for log in generator if getattr(log, topic_id).topic == topic)

        topics = [_remove_0x_prefix(HexBytes32(v).strip()) for v in topic_or_topics]
        return (log for log in generator if getattr(log, topic_id).topic in topics)


def _remove_0x_prefix(string: str) -> str:  # sourcery skip: str-prefix-suffix
    if not isinstance(string, str):
        raise TypeError(type(string), string)
    return string[2:] if string[:2] == "0x" else string



import asyncio
import datetime
import logging
import time
from typing import Final, NewType, Union, final

import dank_mids
import eth_retry
from async_lru import alru_cache
from brownie import chain, web3
from cachetools.func import ttl_cache

try:
    from dank_mids.ENVIRONMENT_VARIABLES import GANACHE_FORK
except ImportError:
    from dank_mids._config import GANACHE_FORK

from y.exceptions import NodeNotSynced
from y.networks import CHAINID, NETWORK_NAME, Network
from y.utils.cache import a_sync_ttl_cache, memory
from y.utils.client import get_ethereum_client, get_ethereum_client_async
from y.utils.logging import yLazyLogger


UnixTimestamp = NewType("UnixTimestamp", int)
Timestamp = Union[UnixTimestamp, datetime.datetime]


logger: Final = logging.getLogger(__name__)
log_debug: Final = logger.debug

now: Final = time.time


@final
class NoBlockFound(Exception):
    """
    Raised when no block is found for a specified timestamp because the timestamp is in the future.

    Args:
        timestamp: The timestamp for which no block was found.
    """

    def __init__(self, timestamp: Timestamp):
        super().__init__(f"No block found after timestamp {timestamp}")


@memory.cache()
@yLazyLogger(logger)
@eth_retry.auto_retry
def get_block_timestamp(height: int) -> int:
    """
    Get the timestamp of a block by its height.

    Args:
        height: The block height.

    Returns:
        The timestamp of the block.

    Examples:
        >>> get_block_timestamp(12345678)
        1609459200

    See Also:
        - :func:`get_block_timestamp_async`
        - :func:`last_block_on_date`
    """
    import y._db.utils.utils as db

    if ts := db.get_block_timestamp(height, sync=True):
        return ts
    client = get_ethereum_client()
    if client in ("tg", "erigon") and CHAINID not in (Network.Polygon,):
        # NOTE: polygon erigon does not support this method
        header = web3.manager.request_blocking(f"{client}_getHeaderByNumber", [height])
        ts = int(header.timestamp, 16)
        db.set_block_timestamp(height, ts, sync=True)
        return ts
    return chain[height].timestamp


@a_sync_ttl_cache
@eth_retry.auto_retry
async def get_block_timestamp_async(height: int) -> int:
    """
    Asynchronously get the timestamp of a block by its height.

    Args:
        height: The block height.

    Returns:
        The timestamp of the block.

    Examples:
        >>> await get_block_timestamp_async(12345678)
        1609459200

    See Also:
        - :func:`get_block_timestamp`
        - :func:`get_block_at_timestamp`
    """
    import y._db.utils.utils as db

    if ts := await db.get_block_timestamp(height, sync=False):
        return ts
    client = await get_ethereum_client_async()
    if client in ("tg", "erigon") and CHAINID not in (Network.Polygon,):
        # NOTE: polygon erigon does not support this method
        header = await dank_mids.web3.manager.coro_request(f"{client}_getHeaderByNumber", [height])
        ts = int(header.timestamp, 16)
    else:
        ts = await dank_mids.eth.get_block_timestamp(height)
    db.set_block_timestamp(height, ts)
    return ts


# TODO: deprecate
@memory.cache()
def last_block_on_date(date: Union[str, datetime.date]) -> int:
    """
    Returns the last block on a given date. You can pass either a `datetime.date` object or a date string formatted as 'YYYY-MM-DD'.

    .. warning::
        This function is marked for deprecation and may be removed in future versions. It is recommended to use alternative methods for obtaining block information.

    Args:
        date: The date for which to find the last block.

    Returns:
        The block number of the last block on the given date.

    Raises:
        TypeError: If a `datetime.datetime` object is passed instead of a `datetime.date`.

    Example:
        >>> last_block_on_date('2023-01-01')
        12345678

    See Also:
        - :func:`get_block_timestamp`
        - :func:`get_block_timestamp_async`
    """
    if isinstance(date, datetime.datetime):
        raise TypeError(
            "You can not pass in a `datetime.datetime` object. Please call date() on your input before passing it to this function."
        )
    if not isinstance(date, datetime.date):
        date = datetime.datetime.strptime(date, "%Y-%m-%d").date()

    height = chain.height
    lo, hi = 0, height
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        log_debug("block: %s", str(mid))
        mid_ts = get_block_timestamp(mid)
        mid_date = datetime.date.fromtimestamp(mid_ts)
        log_debug("mid: %s", mid_date)
        log_debug(date)
        if mid_date > date:
            hi = mid
        else:
            lo = mid
    hi = hi - 1
    block = hi if hi != height else None
    log_debug("last %s block on date %s -> %s", NETWORK_NAME, date, block)
    return block


@a_sync_ttl_cache
async def get_block_at_timestamp(timestamp: datetime.datetime) -> int:
    """
    Get the block number just before a specific timestamp.

    This function returns the block number at the given timestamp, which is the block number just before the specified timestamp.

    Args:
        timestamp: The timestamp to find the block for.

    Returns:
        The block number just before the given timestamp.

    Example:
        >>> await get_block_at_timestamp(datetime.datetime(2023, 1, 1))
        12345678

    See Also:
        - :func:`get_block_timestamp`
        - :func:`get_block_timestamp_async`
    """
    import y._db.utils.utils as db

    if block_at_timestamp := await db.get_block_at_timestamp(timestamp):
        return block_at_timestamp

    # TODO: invert this and use this fn inside of closest_block_after_timestamp for backwards compatibility before deprecating closest_block_after_timestamp
    block_after_timestamp = await closest_block_after_timestamp_async(timestamp)
    block_at_timestamp = block_after_timestamp - 1
    db.set_block_at_timestamp(timestamp, block_at_timestamp)
    return block_at_timestamp


def _parse_timestamp(timestamp: Timestamp) -> UnixTimestamp:
    """
    Parse a timestamp into a Unix timestamp.

    Args:
        timestamp: The timestamp to parse.

    Returns:
        The Unix timestamp.

    Raises:
        TypeError: If the input is not a valid timestamp type.

    Examples:
        >>> _parse_timestamp(datetime.datetime(2023, 1, 1))
        1672531200
        >>> _parse_timestamp(1672531200)
        1672531200
    """
    if isinstance(timestamp, datetime.datetime):
        timestamp = int(timestamp.timestamp())
    elif not isinstance(timestamp, int):
        raise TypeError("You may only pass in a unix timestamp or a datetime object.")
    return UnixTimestamp(timestamp)


# yLazyLogger(logger)
def closest_block_after_timestamp(
    timestamp: Timestamp, wait_for_block_if_needed: bool = False
) -> int:
    """
    Get the closest block after a given timestamp.

    Args:
        timestamp: The timestamp to find the closest block after.
        wait_for_block_if_needed: Whether to wait for a block if needed.

    Returns:
        The block number closest after the given timestamp.

    Raises:
        NoBlockFound: If no block is found after the timestamp.

    Example:
        >>> closest_block_after_timestamp(1672531199)
        12345678

    See Also:
        - :func:`get_block_at_timestamp`
        - :func:`get_block_timestamp`
    """
    timestamp = _parse_timestamp(timestamp)
    while wait_for_block_if_needed:
        try:
            return closest_block_after_timestamp(timestamp)
        except NoBlockFound:
            time.sleep(0.2)
    check_node()
    block = _closest_block_after_timestamp_cached(timestamp)
    log_debug("closest %s block after timestamp %s -> %s", NETWORK_NAME, timestamp, block)
    return block


@a_sync_ttl_cache
async def closest_block_after_timestamp_async(
    timestamp: Timestamp, wait_for_block_if_needed: bool = False
) -> int:
    """
    Asynchronously get the closest block after a given timestamp.

    Args:
        timestamp: The timestamp to find the closest block after.
        wait_for_block_if_needed: Whether to wait for a block if needed.

    Returns:
        The block number closest after the given timestamp.

    Raises:
        NoBlockFound: If no block is found after the timestamp.

    Example:
        >>> await closest_block_after_timestamp_async(1672531199)
        12345678

    See Also:
        - :func:`get_block_at_timestamp`
        - :func:`get_block_timestamp_async`
    """
    timestamp = _parse_timestamp(timestamp)
    while wait_for_block_if_needed:
        try:
            block_at_ts = await get_block_at_timestamp(
                datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc),
                sync=False,
            )
            return block_at_ts + 1
        except NoBlockFound:
            await asyncio.sleep(0.2)

    await check_node_async()

    height = await dank_mids.eth.block_number
    lo, hi = 0, height
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        if await get_block_timestamp_async(mid) > timestamp:
            hi = mid
        else:
            lo = mid
    if hi == height:
        raise NoBlockFound(timestamp)
    return hi


@memory.cache()
def _closest_block_after_timestamp_cached(timestamp: int) -> int:
    height = chain.height
    lo, hi = 0, height
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        if get_block_timestamp(mid) > timestamp:
            hi = mid
        else:
            lo = mid
    if hi == height:
        raise NoBlockFound(timestamp)
    return hi


@ttl_cache(ttl=300)
@eth_retry.auto_retry
def check_node() -> None:
    """
    Check if the Ethereum node is synced.

    Raises:
        y.exceptions.NodeNotSynced: If the node is not synced.

    Examples:
        >>> check_node()
    """
    if GANACHE_FORK:
        return
    current_time = now()
    node_timestamp = web3.eth.get_block("latest").timestamp
    if current_time - node_timestamp > 5 * 60:
        raise NodeNotSynced(
            f"current time: {current_time}  latest block time: {node_timestamp}  discrepancy: {round((current_time - node_timestamp) / 60, 2)} minutes"
        )


@alru_cache(ttl=300)
@eth_retry.auto_retry
async def check_node_async() -> None:
    """
    Asynchronously check if the Ethereum node is synced.

    Raises:
        y.exceptions.NodeNotSynced: If the node is not synced.

    Examples:
        >>> await check_node_async()
    """
    if GANACHE_FORK:
        return
    current_time = now()
    node_timestamp = await dank_mids.eth.get_block_timestamp("latest")
    if current_time - node_timestamp > 5 * 60:
        raise NodeNotSynced(
            f"current time: {current_time}  latest block time: {node_timestamp}  discrepancy: {round((current_time - node_timestamp) / 60, 2)} minutes"
        )



import logging
from contextlib import suppress
from decimal import Decimal
from typing import Dict, List, Optional, Tuple

import a_sync
from a_sync import cgather
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import chain
from brownie.convert.datatypes import EthAddress
from brownie.exceptions import VirtualMachineError
from typing_extensions import Self
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20
from y.constants import dai, usdc, wbtc, weth
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import (
    Address,
    AddressOrContract,
    AnyAddressType,
    Block,
    Pool,
    UsdPrice,
    UsdValue,
)
from y.exceptions import continue_if_call_reverted
from y.networks import Network
from y.prices import magic
from y.prices.dex.balancer._abc import BalancerABC, BalancerPool
from y.utils.cache import optional_async_diskcache

EXCHANGE_PROXY = {
    Network.Mainnet: "0x3E66B66Fd1d0b02fDa6C811Da9E0547970DB2f21",
}.get(chain.id)

SCALES_TO_TRY = [1.0, 0.5, 0.1]
TOKENOUTS_TO_TRY = [weth, dai, usdc, wbtc]

logger = logging.getLogger(__name__)


async def _calc_out_value(
    token_out: AddressOrContract,
    total_outout: int,
    scale: float,
    block: int,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> float:
    """Calculate the output value for a given token.

    Args:
        token_out: The output token address or contract.
        total_outout: The total output amount.
        scale: The scale factor.
        block: The block number.
        skip_cache: Whether to skip the cache.

    Returns:
        The calculated output value.

    Examples:
        >>> await _calc_out_value(weth, 1000, 1.0, 12345678)
        0.5

    See Also:
        - :func:`y.prices.magic.get_price`
    """
    out_scale, out_price = await cgather(
        ERC20._get_scale_for(token_out),
        magic.get_price(token_out, block, skip_cache=skip_cache, sync=False),
    )
    return (total_outout / out_scale) * float(out_price) / scale


class BalancerV1Pool(BalancerPool):
    """A Balancer V1 Pool."""

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    # @optional_async_diskcache
    async def tokens(self) -> List[ERC20]:
        """Get the list of tokens in the pool.

        Returns:
            A list of :class:`~y.classes.common.ERC20` tokens in the pool.

        Examples:
            >>> pool = BalancerV1Pool("0x1234567890abcdef1234567890abcdef12345678")
            >>> await pool.tokens
            [<ERC20 TKN '0x1234567890abcdef1234567890abcdef12345678'>, ...]

        See Also:
            - :class:`~y.classes.common.ERC20`
        """
        contract = await Contract.coroutine(self.address)
        return [
            ERC20(token, asynchronous=self.asynchronous) for token in await contract.getFinalTokens
        ]

    __tokens__: HiddenMethodDescriptor[Self, List[ERC20]]

    @stuck_coro_debugger
    async def get_tvl(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> Optional[UsdValue]:
        """Get the total value locked (TVL) in the pool.

        Args:
            block: The block number to query.
            skip_cache: Whether to skip the cache.

        Returns:
            The total value locked in the pool, or None if it cannot be determined.

        Examples:
            >>> pool = BalancerV1Pool("0x1234567890abcdef1234567890abcdef12345678")
            >>> await pool.get_tvl()
            123456.78

        See Also:
            - :class:`~y.datatypes.UsdValue`
        """
        token_balances = await self.get_balances(block=block, sync=False)
        good_balances = {
            token: balance
            for token, balance in token_balances.items()
            if await token.price(
                block=block,
                return_None_on_failure=True,
                skip_cache=skip_cache,
                sync=False,
            )
            is not None
        }

        if not good_balances:
            return None

        prices = await ERC20.price.map(
            good_balances,
            block=block,
            return_None_on_failure=True,
            skip_cache=skip_cache,
        ).values()

        # in case we couldn't get prices for all tokens, we can extrapolate from the prices we did get
        good_value = sum(
            balance * Decimal(price) for balance, price in zip(good_balances.values(), prices)
        )

        return good_value / len(good_balances) * len(token_balances)

    @stuck_coro_debugger
    async def get_balances(self, block: Optional[Block] = None) -> Dict[ERC20, Decimal]:
        """Get the balances of tokens in the pool.

        Args:
            block: The block number to query.

        Returns:
            A dictionary mapping :class:`~y.classes.common.ERC20` tokens to their balances.

        Examples:
            >>> pool = BalancerV1Pool("0x1234567890abcdef1234567890abcdef12345678")
            >>> await pool.get_balances()
            {<ERC20 TKN '0x1234567890abcdef1234567890abcdef12345678'>: Decimal('1000'), ...}

        See Also:
            - :class:`~y.classes.common.ERC20`
        """
        return await a_sync.map(self.get_balance, self.__tokens__, block=block or "latest")

    @stuck_coro_debugger
    async def get_balance(self, token: AnyAddressType, block: Block) -> Decimal:
        """Get the balance of a specific token in the pool.

        Args:
            token: The token address.
            block: The block number to query.

        Returns:
            The balance of the token in the pool.

        Examples:
            >>> pool = BalancerV1Pool("0x1234567890abcdef1234567890abcdef12345678")
            >>> await pool.get_balance("0xabcdefabcdefabcdefabcdefabcdefabcdef")
            Decimal('1000')
        """
        balance, scale = await cgather(
            self.check_liquidity(str(token), block, sync=False),
            ERC20._get_scale_for(token),
        )
        return Decimal(balance) / scale

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=10_000, ram_cache_ttl=10 * 60)
    async def check_liquidity(self, token: Address, block: Block) -> int:
        """Check the liquidity of a specific token in the pool.

        Args:
            token: The token address.
            block: The block number to query.

        Returns:
            The liquidity of the token in the pool.

        Examples:
            >>> pool = BalancerV1Pool("0x1234567890abcdef1234567890abcdef12345678")
            >>> await pool.check_liquidity("0xabcdefabcdefabcdefabcdefabcdefabcdef")
            1000
        """
        if block < await self.deploy_block(sync=False):
            return 0
        contract = await Contract.coroutine(self.address)
        try:
            return await contract.getBalance.coroutine(token, block_identifier=block)
        except Exception as e:
            # the pool was not yet finalized at this block
            # NOTE: does this happen for any pool except YLA? tbd...
            if "NOT_BOUND" in str(e):
                return 0
            elif e.args and str(e.args[0]) == "execution reverted":
                # we only want to continue on exact match of the original (no extra context added) exception
                return 0
            raise


class BalancerV1(BalancerABC[BalancerV1Pool]):
    """A Balancer V1 instance."""

    _pool_type = BalancerV1Pool

    _check_methods = (
        "getCurrentTokens()(address[])",
        "getTotalDenormalizedWeight()(uint)",
        "totalSupply()(uint)",
    )

    def __init__(self, *, asynchronous: bool = False) -> None:
        """Initialize a BalancerV1 instance.

        Args:
            asynchronous: Whether to use asynchronous operations.

        Examples:
            >>> balancer = BalancerV1(asynchronous=True)
        """
        super().__init__()
        self.asynchronous = asynchronous
        self.exchange_proxy = Contract(EXCHANGE_PROXY) if EXCHANGE_PROXY else None

    @stuck_coro_debugger
    async def get_token_price(
        self,
        token_address: AddressOrContract,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[UsdPrice]:
        """Get the price of a token in the pool.

        Args:
            token_address: The token address or contract.
            block: The block number to query.
            skip_cache: Whether to skip the cache.

        Returns:
            The price of the token in USD, or None if it cannot be determined.

        Examples:
            >>> balancer = BalancerV1(asynchronous=True)
            >>> await balancer.get_token_price("0xabcdefabcdefabcdefabcdefabcdefabcdef")
            1.23

        See Also:
            - :class:`~y.datatypes.UsdPrice`
        """
        if block is not None and block < await contract_creation_block_async(
            self.exchange_proxy, True
        ):
            return None
        for scale in SCALES_TO_TRY:
            # Can we get an output if we try smaller size? try consecutively smaller
            if output := await self.get_some_output(
                token_address, block=block, scale=scale, sync=False
            ):
                return await _calc_out_value(*output, scale, block=block, skip_cache=skip_cache)

    @stuck_coro_debugger
    async def check_liquidity_against(
        self,
        token_in: AddressOrContract,
        token_out: AddressOrContract,
        scale: int = 1,
        block: Optional[Block] = None,
    ) -> Optional[int]:
        """Check the liquidity of a token against another token in the pool.

        Args:
            token_in: The input token address or contract.
            token_out: The output token address or contract.
            scale: The scale factor.
            block: The block number to query.

        Returns:
            The total output amount, or None if it cannot be determined.

        Examples:
            >>> balancer = BalancerV1(asynchronous=True)
            >>> await balancer.check_liquidity_against("0xabcdefabcdefabcdefabcdefabcdefabcdef", "0x1234567890abcdef1234567890abcdef12345678")
            1000
        """
        amount_in = await ERC20._get_scale_for(token_in) * scale
        with suppress(ValueError, VirtualMachineError, ContractLogicError):
            # across various dep versions we get these various excs
            view_split_exact_in = await self.exchange_proxy.viewSplitExactIn.coroutine(
                token_in,
                token_out,
                amount_in,
                32,  # NOTE: 32 is max
                block_identifier=block,
            )
            return view_split_exact_in["totalOutput"]

    @stuck_coro_debugger
    async def get_some_output(
        self, token_in: AddressOrContract, scale: int = 1, block: Optional[Block] = None
    ) -> Optional[Tuple[EthAddress, int]]:
        """Get some output for a given input token.

        Args:
            token_in: The input token address or contract.
            scale: The scale factor.
            block: The block number to query.

        Returns:
            A tuple containing the output token address and the total output amount, or None if it cannot be determined.

        Examples:
            >>> balancer = BalancerV1(asynchronous=True)
            >>> await balancer.get_some_output("0xabcdefabcdefabcdefabcdefabcdefabcdef")
            ('0x1234567890abcdef1234567890abcdef12345678', 1000)
        """
        for token_out in TOKENOUTS_TO_TRY:
            if output := await self.check_liquidity_against(
                token_in, token_out, block=block, scale=scale, sync=False
            ):
                return token_out, output

    @stuck_coro_debugger
    async def check_liquidity(
        self, token: Address, block: Block, ignore_pools: Tuple[Pool, ...] = ()
    ) -> int:
        """Check the liquidity of a token in the pool.

        Args:
            token: The token address.
            block: The block number to query.
            ignore_pools: A tuple of pools to ignore.

        Returns:
            The liquidity of the token in the pool.

        Examples:
            >>> balancer = BalancerV1(asynchronous=True)
            >>> await balancer.check_liquidity("0xabcdefabcdefabcdefabcdefabcdefabcdef", 12345678)
            1000
        """
        pools = [pool for pool in pools if pool not in ignore_pools]
        return (
            await BalancerV1Pool.check_liquidity.max(pools, token=token, block=block, sync=False)
            if pools
            else 0
        )



import logging
from typing import Optional

from y import magic
from y.datatypes import AnyAddressType, Block, UsdPrice

logger = logging.getLogger(__name__)


def get_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    """
    Retrieve the USD price of a given token at a specific block.

    This function serves as a wrapper around `y.magic.get_price` and provides a warning
    about the migration from `ypricemagic` to `y`. The `y` module offers the same functionality
    with additional features.

    .. warning::
        `ypricemagic` is in the process of being migrated to `y`.
        `y` can do all of the same old stuff you expect, plus some new stuff you may or may not need.
        This method still works for now, but will be removed soon.
        Please update your scripts to use `y.magic.get_price(token, block)`.

    Args:
        token: The address of the token to get the price for.
        block: The block number to get the price at. If not provided, the latest block is used.

    Examples:
        >>> from ypricemagic.magic import get_price
        >>> price = get_price("0x6B175474E89094C44Da98b954EedeAC495271d0F")
        >>> print(price)

        >>> price_at_block = get_price("0x6B175474E89094C44Da98b954EedeAC495271d0F", block=12345678)
        >>> print(price_at_block)

    See Also:
        - :func:`y.magic.get_price` for the recommended function to use.
    """
    logger.warning(
        "ypricemagic is in the process of being migrated to y."
        "y can do all of the same old stuff you expect, plus some new stuff you may or may not need."
        "This method still works for now, but will be removed soon."
        "Please update your scripts to use `y.magic.get_price(token, block)`."
    )
    return magic.get_price(token, block)



from typing import Optional

from brownie import Contract as _Contract
from brownie import chain

from y.contracts import Contract
from y.interfaces.ERC20 import ERC20ABI
from y.networks import Network


CHAINID = chain.id
"""
The chainid for the connected rpc.
"""

NETWORK_NAME = Network.name(CHAINID)
"""
The name of the connected network.
"""

EEE_ADDRESS = "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE"
"""
The address used to represent the native token (e.g., ETH on Ethereum, AVAX on Avalanche, etc.) in various DeFi protocols.
"""

sushi: Optional[Contract] = None
"""
A placeholder for the Sushi token contract, which may be set depending on the network.
"""

if CONNECTED_TO_MAINNET := CHAINID == Network.Mainnet:
    weth = Contract("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2")
    """Wrapped Ether (WETH) contract on Ethereum Mainnet."""

    usdc = Contract("0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48")
    """USD Coin (USDC) contract on Ethereum Mainnet."""

    dai = Contract("0x6B175474E89094C44Da98b954EedeAC495271d0F")
    """Dai Stablecoin (DAI) contract on Ethereum Mainnet."""

    wbtc = Contract("0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599")
    """Wrapped Bitcoin (WBTC) contract on Ethereum Mainnet."""

    usdt = Contract("0xdAC17F958D2ee523a2206206994597C13D831ec7")
    """Tether USD (USDT) contract on Ethereum Mainnet."""

    sushi = Contract("0x6B3595068778DD592e39A122f4f5a5cF09C90fE2")
    """SushiToken (SUSHI) contract on Ethereum Mainnet."""

elif CHAINID == Network.BinanceSmartChain:
    weth = Contract("0x2170Ed0880ac9A755fd29B2688956BD959F933F8")
    """Wrapped Ether (WETH) contract on Binance Smart Chain."""

    usdc = Contract("0x8ac76a51cc950d9822d68b83fe1ad97b32cd580d")
    """USD Coin (USDC) contract on Binance Smart Chain."""

    dai = Contract("0x1af3f329e8be154074d8769d1ffa4ee058b1dbc3")
    """Dai Stablecoin (DAI) contract on Binance Smart Chain."""

    wbtc = Contract("0x2ccb7c8c51e55c2364b555ff6e6e3f7246499e16")
    """Wrapped Bitcoin (WBTC) contract on Binance Smart Chain."""

    usdt = Contract("0x55d398326f99059ff775485246999027b3197955")
    """Tether USD (USDT) contract on Binance Smart Chain."""

    wbnb = Contract("0xbb4CdB9CBd36B01bD1cBaEBF2De08d9173bc095c")
    """Wrapped BNB (WBNB) contract on Binance Smart Chain."""

    cake = Contract("0x0e09fabb73bd3ade0a17ecc321fd13a19e81ce82")
    """PancakeSwap Token (CAKE) contract on Binance Smart Chain."""

elif CHAINID == Network.xDai:
    weth = Contract("0x6A023CCd1ff6F2045C3309768eAd9E68F978f6e1")
    """Wrapped Ether (WETH) contract on xDai Chain."""

    wbtc = Contract("0x8e5bBbb09Ed1ebdE8674Cda39A0c169401db4252")
    """Wrapped Bitcoin (WBTC) contract on xDai Chain."""

    dai = Contract("0x44fA8E6f47987339850636F88629646662444217")
    """Dai Stablecoin (DAI) contract on xDai Chain."""

    usdc = Contract("0xDDAfbb505ad214D7b80b1f830fcCc89B60fb7A83")
    """USD Coin (USDC) contract on xDai Chain."""

    usdt = Contract("0x4ECaBa5870353805a9F068101A40E0f32ed605C6")
    """Tether USD (USDT) contract on xDai Chain."""

elif CHAINID == Network.Polygon:
    weth = Contract("0x7ceb23fd6bc0add59e62ac25578270cff1b9f619")
    """Wrapped Ether (WETH) contract on Polygon."""

    usdc = Contract("0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174")
    """USD Coin (USDC) contract on Polygon."""

    dai = Contract("0x8f3Cf7ad23Cd3CaDbD9735AFf958023239c6A063")
    """Dai Stablecoin (DAI) contract on Polygon."""

    wbtc = Contract("0x1bfd67037b42cf73acf2047067bd4f2c47d9bfd6")
    """Wrapped Bitcoin (WBTC) contract on Polygon."""

    usdt = Contract("0xc2132D05D31c914a87C6611C10748AEb04B58e8F")
    """Tether USD (USDT) contract on Polygon."""

    wmatic = Contract("0x0d500B1d8E8eF31E21C99d1Db9A6444d3ADf1270")
    """Wrapped Matic (WMATIC) contract on Polygon."""

elif CHAINID == Network.Fantom:
    weth = Contract("0x74b23882a30290451A17c44f4F05243b6b58C76d")
    """Wrapped Ether (WETH) contract on Fantom."""

    usdc = Contract("0x04068DA6C83AFCFA0e13ba15A6696662335D5B75")
    """USD Coin (USDC) contract on Fantom."""

    dai = Contract("0x8D11eC38a3EB5E956B052f67Da8Bdc9bef8Abf3E")
    """Dai Stablecoin (DAI) contract on Fantom."""

    wbtc = Contract("0x321162Cd933E2Be498Cd2267a90534A804051b11")
    """Wrapped Bitcoin (WBTC) contract on Fantom."""

    usdt = Contract("0x049d68029688eAbF473097a2fC38ef61633A3C7A")  # fusdt
    """Fantom Tether USD (FUSDT) contract on Fantom."""

    wftm = Contract("0x21be370D5312f44cB42ce377BC9b8a0cEF1A4C83")
    """Wrapped Fantom (WFTM) contract on Fantom."""

elif CHAINID == Network.Arbitrum:
    weth = Contract("0x82aF49447D8a07e3bd95BD0d56f35241523fBab1")
    """Wrapped Ether (WETH) contract on Arbitrum."""
    # NOTE: for some reason wbtc unverified on arbi

    wbtc = _Contract.from_abi("wbtc", "0x2f2a2543B76A4166549F7aaB2e75Bef0aefC5B0f", ERC20ABI)
    """Wrapped Bitcoin (WBTC) contract on Arbitrum."""

    dai = Contract("0xDA10009cBd5D07dd0CeCc66161FC93D7c9000da1")
    """Dai Stablecoin (DAI) contract on Arbitrum."""

    usdc = Contract("0xFF970A61A04b1cA14834A43f5dE4533eBDDB5CC8")
    """USD Coin (USDC) contract on Arbitrum."""

    usdt = Contract("0xFd086bC7CD5C481DCC9C85ebE478A1C0b69FCbb9")
    """Tether USD (USDT) contract on Arbitrum."""

elif CHAINID == Network.Avalanche:
    weth = Contract("0x49D5c2BdFfac6CE2BFdB6640F4F80f226bc10bAB")  # weth.e
    """Wrapped Ether (WETH.e) contract on Avalanche."""

    wbtc = Contract("0x50b7545627a5162F82A992c33b87aDc75187B218")
    """Wrapped Bitcoin (WBTC) contract on Avalanche."""

    dai = Contract("0xd586E7F844cEa2F87f50152665BCbc2C279D8d70")
    """Dai Stablecoin (DAI) contract on Avalanche."""

    usdc = Contract("0xA7D7079b0FEaD91F3e65f86E8915Cb59c1a4C664")  # usdc.e
    """USD Coin (USDC.e) contract on Avalanche."""

    usdt = Contract("0x9702230A8Ea53601f5cD2dc00fDBc13d4dF4A8c7")
    """Tether USD (USDT) contract on Avalanche."""

elif CHAINID == Network.Moonriver:
    weth = Contract("0x639A647fbe20b6c8ac19E48E2de44ea792c62c5C")
    """Wrapped Ether (WETH) contract on Moonriver."""

    wbtc = Contract("0x6aB6d61428fde76768D7b45D8BFeec19c6eF91A8")
    """Wrapped Bitcoin (WBTC) contract on Moonriver."""

    dai = Contract("0x80A16016cC4A2E6a2CACA8a4a498b1699fF0f844")
    """Dai Stablecoin (DAI) contract on Moonriver."""

    usdc = Contract("0xE3F5a90F9cb311505cd691a46596599aA1A0AD7D")
    """USD Coin (USDC) contract on Moonriver."""

    usdt = Contract("0xB44a9B6905aF7c801311e8F4E76932ee959c663C")
    """Tether USD (USDT) contract on Moonriver."""

elif CHAINID == Network.Heco:
    weth = Contract("0x64FF637fB478863B7468bc97D30a5bF3A428a1fD")
    """Wrapped Ether (WETH) contract on HECO Chain."""

    wbtc = Contract("0x70D171d269D964d14aF9617858540061e7bE9EF1")
    """Wrapped Bitcoin (WBTC) contract on HECO Chain."""

    dai = Contract("0x3D760a45D0887DFD89A2F5385a236B29Cb46ED2a")
    """Dai Stablecoin (DAI) contract on HECO Chain."""

    usdc = Contract("0x9362Bbef4B8313A8Aa9f0c9808B80577Aa26B73B")
    """USD Coin (USDC) contract on HECO Chain."""

    usdt = Contract("0xa71EdC38d189767582C38A3145b5873052c3e47a")
    """Tether USD (USDT) contract on HECO Chain."""

elif CHAINID == Network.Harmony:
    weth = Contract("0x6983D1E6DEf3690C4d616b13597A09e6193EA013")
    """Wrapped Ether (WETH) contract on Harmony."""

    wbtc = Contract("0x3095c7557bCb296ccc6e363DE01b760bA031F2d9")
    """Wrapped Bitcoin (WBTC) contract on Harmony."""

    dai = Contract("0xEf977d2f931C1978Db5F6747666fa1eACB0d0339")
    """Dai Stablecoin (DAI) contract on Harmony."""

    usdc = Contract("0x985458E523dB3d53125813eD68c274899e9DfAb4")
    """USD Coin (USDC) contract on Harmony."""

    usdt = Contract("0x3C2B8Be99c50593081EAA2A724F0B8285F5aba8f")
    """Tether USD (USDT) contract on Harmony."""

elif CHAINID == Network.Aurora:
    weth = Contract("0xC9BdeEd33CD01541e1eeD10f90519d2C06Fe3feB")
    """Wrapped Ether (WETH) contract on Aurora."""

    wbtc = _Contract.from_abi(
        "ERC20 [forced]", "0xF4eB217Ba2454613b15dBdea6e5f22276410e89e", ERC20ABI
    )
    """Wrapped Bitcoin (WBTC) contract on Aurora."""

    dai = _Contract.from_abi(
        "ERC20 [forced]", "0xe3520349F477A5F6EB06107066048508498A291b", ERC20ABI
    )
    """Dai Stablecoin (DAI) contract on Aurora."""

    usdc = _Contract.from_abi(
        "ERC20 [forced]", "0xB12BFcA5A55806AaF64E99521918A4bf0fC40802", ERC20ABI
    )
    """USD Coin (USDC) contract on Aurora."""

    usdt = _Contract.from_abi(
        "ERC20 [forced]", "0x4988a896b1227218e4A686fdE5EabdcAbd91571f", ERC20ABI
    )
    """Tether USD (USDT) contract on Aurora."""

elif CHAINID == Network.Cronos:
    weth = Contract("0xe44Fd7fCb2b1581822D0c862B68222998a0c299a")
    """Wrapped Ether (WETH) contract on Cronos."""

    wbtc = Contract("0x062E66477Faf219F25D27dCED647BF57C3107d52")
    """Wrapped Bitcoin (WBTC) contract on Cronos."""

    dai = Contract("0xF2001B145b43032AAF5Ee2884e456CCd805F677D")
    """Dai Stablecoin (DAI) contract on Cronos."""

    usdc = Contract("0xc21223249CA28397B4B6541dfFaEcC539BfF0c59")
    """USD Coin (USDC) contract on Cronos."""

    usdt = Contract("0x66e428c3f67a68878562e79A0234c1F83c208770")
    """Tether USD (USDT) contract on Cronos."""

elif CHAINID == Network.Optimism:
    weth = Contract("0x4200000000000000000000000000000000000006")
    """Wrapped Ether (WETH) contract on Optimism."""

    wbtc = Contract("0x68f180fcCe6836688e9084f035309E29Bf0A2095")
    """Wrapped Bitcoin (WBTC) contract on Optimism."""

    dai = Contract("0xDA10009cBd5D07dd0CeCc66161FC93D7c9000da1")
    """Dai Stablecoin (DAI) contract on Optimism."""

    usdc = Contract("0x7F5c764cBc14f9669B88837ca1490cCa17c31607")
    """USD Coin (USDC) contract on Optimism."""

    usdt = Contract("0x94b008aA00579c1307B0EF2c499aD98a8ce58e58")
    """Tether USD (USDT) contract on Optimism."""

elif CHAINID == Network.Base:
    dai = Contract("0x50c5725949A6F0c72E6C4a641F24049A917DB0Cb")
    """Dai Stablecoin (DAI) contract on Base."""

    weth = Contract("0x4200000000000000000000000000000000000006")
    """Wrapped Ether (WETH) contract on Base."""

    usdc = Contract("0xd9aAEc86B65D86f6A7B5B1b0c42FFA531710b6CA")  # usdbc
    """USD Coin (USDC) contract on Base."""

    usdt = Contract("0x4A3A6Dd60A34bB2Aba60D73B4C88315E9CeB6A3D")
    """Tether USD (USDT) contract on Base."""

    wbtc = Contract("0x77852193BD608A518dd7b7C2f891A1d02ceeB4d4")
    """
    Wrapped Bitcoin (WBTC) contract on Base.

    Note:
        This is a temporary placeholder and may not represent the actual WBTC contract on Base.
    """

else:
    weth, dai, wbtc, usdc, usdt = None, None, None, None, None

STABLECOINS = {
    Network.Mainnet: {
        "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48": "usdc",
        "0x0000000000085d4780B73119b644AE5ecd22b376": "tusd",
        "0x6B175474E89094C44Da98b954EedeAC495271d0F": "dai",
        "0xdAC17F958D2ee523a2206206994597C13D831ec7": "usdt",
        "0x57Ab1E02fEE23774580C119740129eAC7081e9D3": "susd_old",
    },
    Network.BinanceSmartChain: {
        "0x8ac76a51cc950d9822d68b83fe1ad97b32cd580d": "usdc",
        "0x1af3f329e8be154074d8769d1ffa4ee058b1dbc3": "dai",
        "0x55d398326f99059ff775485246999027b3197955": "usdt",
        "0xe9e7CEA3DedcA5984780Bafc599bD69ADd087D56": "busd",
        "0x23396cF899Ca06c4472205fC903bDB4de249D6fC": "wust",
        "0x14016E85a25aeb13065688cAFB43044C2ef86784": "tusd",
    },
    Network.Polygon: {
        "0x3c499c542cEF5E3811e1192ce70d8cC03d5c3359": "usdc",
        "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174": "usdc.e",
        "0x8f3Cf7ad23Cd3CaDbD9735AFf958023239c6A063": "dai",
        "0xc2132D05D31c914a87C6611C10748AEb04B58e8F": "usdt",
        "0xE840B73E5287865EEc17d250bFb1536704B43B21": "musd",
    },
    Network.Fantom: {
        "0x04068DA6C83AFCFA0e13ba15A6696662335D5B75": "usdc",
        "0x8D11eC38a3EB5E956B052f67Da8Bdc9bef8Abf3E": "dai",
        "0x049d68029688eAbF473097a2fC38ef61633A3C7A": "fusdt",
        "0x9879aBDea01a879644185341F7aF7d8343556B7a": "tusd",
    },
    Network.Arbitrum: {
        "0xDA10009cBd5D07dd0CeCc66161FC93D7c9000da1": "dai",
        "0xFF970A61A04b1cA14834A43f5dE4533eBDDB5CC8": "usdc",
        "0xFd086bC7CD5C481DCC9C85ebE478A1C0b69FCbb9": "usdt",
    },
    Network.Avalanche: {
        "0xde3A24028580884448a5397872046a019649b084": "usdt",
        "0xc7198437980c041c805A1EDcbA50c1Ce5db95118": "usdt.e",
        "0xB97EF9Ef8734C71904D8002F8b6Bc66Dd9c48a6E": "usdc",
        "0xA7D7079b0FEaD91F3e65f86E8915Cb59c1a4C664": "usdc.e",
        "0xbA7dEebBFC5fA1100Fb055a87773e1E99Cd3507a": "dai",
        "0xd586E7F844cEa2F87f50152665BCbc2C279D8d70": "dai.e",
        "0x130966628846BFd36ff31a822705796e8cb8C18D": "mim",
        "0xaEb044650278731Ef3DC244692AB9F64C78FfaEA": "busd",
    },
    Network.Heco: {
        "0xa71EdC38d189767582C38A3145b5873052c3e47a": "usdt",
        "0x9362Bbef4B8313A8Aa9f0c9808B80577Aa26B73B": "usdc",
        "0x3D760a45D0887DFD89A2F5385a236B29Cb46ED2a": "dai",
    },
    Network.Harmony: {
        "0x3C2B8Be99c50593081EAA2A724F0B8285F5aba8f": "usdt",
        "0x985458E523dB3d53125813eD68c274899e9DfAb4": "usdc",
        "0xEf977d2f931C1978Db5F6747666fa1eACB0d0339": "dai",
    },
    Network.Cronos: {
        "0x66e428c3f67a68878562e79A0234c1F83c208770": "usdt",
        "0xc21223249CA28397B4B6541dfFaEcC539BfF0c59": "usdc",
        "0xF2001B145b43032AAF5Ee2884e456CCd805F677D": "dai",
    },
    Network.Optimism: {
        "0xDA10009cBd5D07dd0CeCc66161FC93D7c9000da1": "dai",
        "0x7F5c764cBc14f9669B88837ca1490cCa17c31607": "usdc",
        "0x94b008aA00579c1307B0EF2c499aD98a8ce58e58": "usdt",
    },
    Network.Base: {
        "0xd9aAEc86B65D86f6A7B5B1b0c42FFA531710b6CA": "usdbc",
        "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913": "usdc",
    },
}.get(CHAINID, {})
"""
A dictionary mapping network IDs to stablecoin contract addresses and their corresponding symbols.

Each network has a set of stablecoin addresses with their associated symbols. This allows for easy lookup of stablecoin contracts on different networks. If the current `chain.id` is not recognized, it defaults to an empty dictionary.

Example:
    >>> STABLECOINS.get(Network.Mainnet, {}).get("0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48")
    'usdc'

See Also:
    - :class:`~y.networks.Network` for network ID definitions.
"""

WRAPPED_GAS_COIN = {
    Network.Mainnet: "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
    Network.BinanceSmartChain: "0xbb4CdB9CBd36B01bD1cBaEBF2De08d9173bc095c",
    Network.Polygon: "0x0d500B1d8E8eF31E21C99d1Db9A6444d3ADf1270",
    Network.Fantom: "0x21be370D5312f44cB42ce377BC9b8a0cEF1A4C83",
    Network.Arbitrum: "0x82aF49447D8a07e3bd95BD0d56f35241523fBab1",
    Network.Avalanche: "0xB31f66AA3C1e785363F0875A1B74E27b85FD66c7",
    Network.Heco: "0x5545153CCFcA01fbd7Dd11C0b23ba694D9509A6F",
    Network.Harmony: "0xcF664087a5bB0237a0BAd6742852ec6c8d69A27a",
    Network.Cronos: "0x5C7F8A570d578ED84E63fdFA7b1eE72dEae1AE23",
    Network.Optimism: "0x4200000000000000000000000000000000000006",
    Network.Base: "0x4200000000000000000000000000000000000006",
}.get(CHAINID)
"""
The address of the wrapped version of the native token on the active network.

For example, on Ethereum Mainnet, `WRAPPED_GAS_COIN` == the WETH address. On Fantom, it is equal to the WFTM address. And so on.

Example:
    >>> WRAPPED_GAS_COIN
    '0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2'  # On Ethereum Mainnet

See Also:
    - :class:`~y.networks.Network` for network ID definitions.
"""
